import os
import json
import time
import hmac
import random
import asyncio
import hashlib

from urllib.parse import urljoin
from aiolimiter import AsyncLimiter


import httpx
import websockets


RATE_LIMITS = {
    # order methods
    (
        'private/create-order',
        'private/cancel-order',
        'private/cancel-all-orders',
        'private/margin/create-order',
        'private/margin/cancel-order',
        'private/margin/cancel-all-orders',
    ): (14, 0.1),

    # order detail methods
    (
        'private/get-order-detail',
        'private/margin/get-order-detail',
    ): (29, 0.1),

    # general trade methods
    (
        'private/get-trades',
        'private/margin/get-trades',
        'private/get-order-history',
        'private/margin/get-order-history'
    ): (1, 1)
}


class ApiError(Exception):
    pass


class ApiListenAsyncIterable:
    def __init__(self, api, ws, channels, sign):
        self.api = api
        self.ws = ws
        self.channels = channels

        self.sign = sign
        self.sub_data_sent = False
        self.auth_sent = False
        self.connected = False
        self.auth_data = None

    async def connect(self):
        self.auth_data = {}
        if self.sign:
            self.auth_data = self.api.sign('public/auth', self.auth_data)

        # sleep because too many requests from docs
        await asyncio.sleep(1)

        self.connected = True

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.connected:
            await asyncio.sleep(1)

        if not self.sub_data_sent:
            sub_data = {
                'id': random.randint(1000, 10000),
                "method": "subscribe",
                'params': {"channels": self.channels},
                "nonce": int(time.time())
            }

        # [0] if not sign start connection with subscription
        if not self.sub_data_sent and not self.sign:
            await self.ws.send(json.dumps(sub_data))
            self.sub_data_sent = True

        data = await self.ws.recv()

        if data:
            data = json.loads(data)
            result = data.get('result')

            # [1] send heartbeat to keep connection alive
            if data['method'] == 'public/heartbeat':
                await self.ws.send(json.dumps({
                    'id': data['id'],
                    'method': 'public/respond-heartbeat'
                }))
            elif self.sub_data_sent and result:
                # [4] consume data
                return result

            # [2] sign auth request to listen private methods
            if self.sign and not self.auth_sent:
                await self.ws.send(json.dumps(self.auth_data))
                self.auth_sent = True

            # [3] subscribe to channels
            if (
                    data['method'] == 'public/auth' and
                    data['code'] == 0 and not self.sub_data_sent
            ):
                await self.ws.send(json.dumps(sub_data))
                self.sub_data_sent = True



class ApiProvider:
    """Provides HTTP-api requests and websocket requests."""
    def __init__(
            self, *, api_key='', api_secret='', from_env=False,
            auth_required=True, timeout=25, retries=6,
            root_url='https://api.crypto.com/v2/',
            ws_root_url='wss://stream.crypto.com/v2/', logger=None):

        self.api_key = api_key
        self.api_secret = api_secret
        self.root_url = root_url
        self.ws_root_url = ws_root_url
        self.timeout = timeout
        self.retries = retries
        self.last_request_path = ''

        self.rate_limiters = {}

        for urls in RATE_LIMITS:
            for url in urls:
                self.rate_limiters[url] = AsyncLimiter(*RATE_LIMITS[urls])

        # limits for not matched methods
        self.general_private_limit = AsyncLimiter(3, 0.1)
        self.general_public_limit = AsyncLimiter(100, 1)

        if not auth_required:
            return

        if from_env:
            self.read_keys_from_env()

        if not self.api_key or not self.api_secret:
            raise ValueError('Provide api_key and api_secret')

    def read_keys_from_env(self):
        self.api_key = os.environ.get('CRYPTOCOM_API_KEY', '')
        if not self.api_key:
            raise ValueError('Provide CRYPTOCOM_API_KEY env value')
        self.api_secret = os.environ.get('CRYPTOCOM_API_SECRET', '')
        if not self.api_secret:
            raise ValueError('Provide CRYPTOCOM_API_SECRET env value')

    def sign(self, path, data):
        data = data or {}
        data['method'] = path

        sign_time = int(time.time() * 1000)
        data.update({'nonce': sign_time, 'api_key': self.api_key})

        data['id'] = random.randint(1000, 10000)
        data_params = data.get('params', {})
        params = ''.join(
            f'{key}{data_params[key]}'
            for key in sorted(data_params)
        )

        payload = f"{path}{data['id']}" \
            f"{self.api_key}{params}{data['nonce']}"

        data['sig'] = hmac.new(
            self.api_secret.encode('utf-8'),
            payload.encode('utf-8'),
            hashlib.sha256
        ).hexdigest()
        return data

    def get_limiter(self, path):
        if path in self.rate_limiters.keys():
            return self.rate_limiters[path]
        else:
            if path.startswith('private'):
                return self.general_private_limit
            elif path.startswith('public'):
                return self.general_public_limit
            else:
                raise ApiError(f'Wrong path: {path}')

    async def request(self, method, path, params=None, data=None, sign=False):
        original_data = data
        limiter = self.get_limiter(path)

        for count in range(self.retries + 1):
            client = httpx.AsyncClient(
                timeout=httpx.Timeout(timeout=self.timeout))
            if sign:
                data = self.sign(path, original_data)
            try:
                async with limiter:
                    resp = await client.request(
                        method, urljoin(self.root_url, path),
                        params=params, json=data,
                        headers={'content-type': 'application/json'}
                    )
                    resp_json = resp.json()
                    if resp.status_code != 200:
                        raise ApiError(
                            f"Error: {resp_json}. "
                            f"Status: {resp.status_code}. Json params: {data}")
            except httpx.ConnectError:
                raise ApiError(f"Cannot connect to host {self.root_url}")
            except asyncio.TimeoutError as exc:
                if count == self.retries:
                    raise ApiError(
                        f"Timeout error, retries: {self.retries}. "
                        f"Code: {resp.status_code}. Data: {data}"
                    ) from exc
                continue
            except json.JSONDecodeError:
                if resp.status_code == 429:
                    continue
                text = await resp.text
                raise ApiError(
                    f"Can't decode json, content: {text}. "
                    f"Code: {resp.status_code}")
            finally:
                await client.aclose()

            if resp_json['code'] == 0:
                result = resp_json.get('result', {})
                return result.get('data', {}) or result

            if count == self.retries:
                raise ApiError(
                    f"System error, retries: {self.retries}. "
                    f"Code: {resp.status_code}. Json: {resp_json}. "
                    f"Data: {data}"
                )

    async def get(self, path, params=None, sign=False):
        return await self.request('get', path, params=params, sign=sign)

    async def post(self, path, data=None, sign=True):
        return await self.request('post', path, data=data, sign=sign)

    async def listen(self, url, *channels, sign=False):
        url = urljoin(self.ws_root_url, url)
        async for ws in websockets.connect(url, open_timeout=self.timeout):
            try:
                dataiterator = ApiListenAsyncIterable(self, ws, channels, sign)
                async for data in dataiterator:
                    if data:
                        yield data
            except websockets.ConnectionClosed:
                continue



def round_up(value: float, precision: int) -> float:
    pwr = 10 ** precision
    rounded = round((value * pwr) / pwr, precision)
    if rounded < value:
        rounded = round(
            (value * pwr + 1) / pwr, precision)
    return rounded


def round_down(value: float, precision: int) -> float:
    pwr = 10 ** precision
    return round((value * pwr) / pwr, precision)



import asyncio
import platform

from .structs import (
    Order, OrderSide, OrderStatus, OrderType, Pair, Period, Candle,
    MarketTrade, Coin, PrivateTrade, Timeframe, DepositStatus, Deposit,
    WithdrawalStatus, Withdrawal
)
from .market import Exchange
from .private import Account
from .api import ApiError, ApiProvider
from . import pairs, coins

if platform.system() == 'Windows':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

__all__ = [
    'Order', 'OrderSide', 'OrderStatus', 'OrderType',
    'pairs', 'Pair', 'coins', 'Coin',
    'Period', 'Candle', 'MarketTrade', 'PrivateTrade',
    'Timeframe', 'Deposit', 'Withdrawal', 'DepositStatus', 'WithdrawalStatus',
    'Exchange', 'Account',
    'ApiError', 'ApiProvider'
]

__version__ = '0.10.0'



import asyncio

from typing import List, Dict

from .api import ApiProvider
from .structs import (
    Pair, MarketTicker, MarketTrade, Period, Candle,
    OrderInBook, OrderBook, OrderSide
)
from . import pairs


class Exchange:
    """Interface to base exchange methods."""
    def __init__(self, api: ApiProvider = None):
        self.api = api or ApiProvider(auth_required=False)
        self.pairs = {pair.name: pair for pair in pairs.all()}

    async def sync_pairs(self):
        """Use this method to sync pairs if you have issues with missing
        pairs in library side."""
        self.pairs = {pair.name: pair for pair in (await self.get_pairs())}

    async def get_pairs(self) -> List[Pair]:
        """List all available market pairs and store to provide pairs info."""
        data = await self.api.get('public/get-instruments')
        return [
            Pair(
                i['instrument_name'],
                price_precision=i['price_decimals'],
                quantity_precision=i['quantity_decimals']
            ) for i in data['instruments']
        ]

    async def get_ticker(self, pair: Pair) -> MarketTicker:
        """Get ticker in for provided pair."""
        data = await self.api.get(
            'public/get-ticker', {'instrument_name': pair.name})
        return MarketTicker.from_api(pair, data)

    async def get_tickers(self) -> Dict[Pair, MarketTicker]:
        """Get tickers in all available markets."""
        data = await self.api.get('public/get-ticker')
        return {
            self.pairs[ticker['i']]: MarketTicker.from_api(
                self.pairs[ticker['i']], ticker
            ) for ticker in data
            if ticker['i'] in self.pairs
        }

    async def get_price(self, pair: Pair) -> float:
        """Get latest price of pair."""
        return (await self.get_ticker(pair)).trade_price

    async def get_trades(self, pair: Pair) -> List[MarketTrade]:
        """Get last 200 trades in a specified market."""
        data = await self.api.get(
            'public/get-trades', {'instrument_name': pair.name})
        return [MarketTrade.from_api(pair, trade) for trade in reversed(data)]

    async def get_orderbook(self, pair: Pair) -> OrderBook:
        """Get the order book for a particular market, depth always 150."""
        data = await self.api.get('public/get-book', {
            'instrument_name': pair.name})
        buys = [
            OrderInBook(*order, pair, OrderSide.BUY)
            for order in data[0]['bids']
        ]
        sells = [
            OrderInBook(*order, pair, OrderSide.SELL)
            for order in reversed(data[0]['asks'])
        ]
        return OrderBook(buys, sells, pair)

    async def get_candles(self, pair: Pair, period: Period) -> List[Candle]:
        data = await self.api.get('public/get-candlestick', {
            'instrument_name': pair.name, 'timeframe': period.value})
        return [Candle.from_api(pair, candle) for candle in data]

    async def listen_candles(
            self, period: Period, *pairs: List[Pair]) -> Candle:
        if not isinstance(period, Period):
            raise ValueError(f'Provide Period enum not {period}')

        channels = [
            f'candlestick.{period}.{pair.name}'
            for pair in pairs
        ]
        prev_time = {}

        async for data in self.api.listen('market', *channels):
            pair = self.pairs[data['instrument_name']]
            for candle in data['data']:
                current_time = int(candle['t'] / 1000)
                if pair not in prev_time or current_time > prev_time[pair]:
                    yield Candle.from_api(pair, candle)
                    prev_time[pair] = current_time

    async def listen_trades(self, *pairs: List[Pair]) -> MarketTrade:
        channels = [f'trade.{pair.name}' for pair in pairs]
        async for data in self.api.listen('market', *channels):
            for trade in data['data']:
                pair = self.pairs[data['instrument_name']]
                yield MarketTrade.from_api(pair, trade)

    async def listen_orderbook(self, *pairs: List[Pair]) -> OrderBook:
        channels = [f'book.{pair.name}.150' for pair in pairs]
        async for data in self.api.listen('market', *channels):
            pair = self.pairs[data['instrument_name']]
            buys = [
                OrderInBook(*order, pair, OrderSide.BUY)
                for order in data['data'][0]['bids']
            ]
            sells = [
                OrderInBook(*order, pair, OrderSide.SELL)
                for order in reversed(data['data'][0]['asks'])
            ]
            yield OrderBook(buys, sells, pair)
